package MySTDIN;
use warnings;
use strict;
use Test::More;
our @E = (0, 0, 0, 0);
sub _check {
    my ($self, $event, $fh) = @_;
    is($event, $main::stdin, 'STDIN event');
    is(fileno($fh), fileno(*STDIN), 'STDIN filehandle');
}
sub IN  { _check(@_); $E[0]++ }
sub OUT { _check(@_); $E[1]++ }
sub HUP { _check(@_); $E[2]++ }
sub ERR { _check(@_); $E[3]++ }


package MySTDOUT;
use warnings;
use strict;
use Test::More;
our @E = (0, 0, 0, 0);
sub _check {
    my ($self, $event, $fh) = @_;
    is($event, $main::stdout, 'STDOUT event');
    is(fileno($fh), fileno(*STDOUT), 'STDOUT filehandle');
}
sub IN  { _check(@_); $E[0]++ }
sub OUT { _check(@_); $E[1]++ }
sub HUP { _check(@_); $E[2]++ }
sub ERR { _check(@_); $E[3]++ }


package Suicide;
use warnings;
use strict;
our $Flag = 0;
sub new {
    my $self = bless {}, 'Suicide';
    $self->{me} = $self;            # I own self
    return $self;
}
sub IN  { $Flag++ }
sub OUT { $Flag++ }
sub HUP {
    my ($self) = @_;
    undef $self->{me};              # kill myself
}

package main;
use warnings;
use strict;
use Test::More tests => 60;
use Test::Exception;
use Test::MockModule;
use Errno qw( EINTR );
use Scalar::Util qw( isweak weaken );

use POWER::Event::Epoll qw( :ALL );
use POWER::Epoll;
use POWER::Utils;

sub loop {
    my ($status, @events) = Epoll_wait(0);
    if ($status) {
#        warn("run_callback(@$_)\n"),
        defined $_->[0] && run_callback(@$_) for @events;
    }
    return $status;
}

my $module = Test::MockModule->new('POWER::Event::Epoll');
$module->mock('epoll_create'    => sub {255});
$module->mock('epoll_add'       => sub {1});
$module->mock('epoll_del'       => sub {1});

our ($stdin, $stdout);
$stdin  = Epoll_add(*STDIN,  'MySTDIN' );
$stdout = Epoll_add(*STDOUT, 'MySTDOUT');

is_deeply( \@MySTDIN::E,  [ 0, 0, 0, 0 ] );
is_deeply( \@MySTDOUT::E, [ 0, 0, 0, 0 ] );

$module->mock('epoll_wait' => sub { return(2, [
    [ fileno(*STDIN),  EPOLLIN  ],
    [ fileno(*STDOUT), EPOLLOUT ],
]) });
is(loop(), 2);
is_deeply( \@MySTDIN::E,  [ 1, 0, 0, 0 ] );
is_deeply( \@MySTDOUT::E, [ 0, 1, 0, 0 ] );
is(loop(), 2);
is_deeply( \@MySTDIN::E,  [ 2, 0, 0, 0 ] );
is_deeply( \@MySTDOUT::E, [ 0, 2, 0, 0 ] );

$module->mock('epoll_wait' => sub { return(1, [
    [ fileno(*STDIN),  EPOLLIN|EPOLLHUP|EPOLLERR ],
]) });
is(loop(), 1);
is_deeply( \@MySTDIN::E,  [ 3, 0, 1, 1 ] );
is_deeply( \@MySTDOUT::E, [ 0, 2, 0, 0 ] );

$module->mock('epoll_wait' => sub { $!=EINTR; return(-1, [
    [ fileno(*STDIN),  EPOLLIN|EPOLLHUP|EPOLLERR ],
]) });
is(loop(), undef);
is_deeply( \@MySTDIN::E,  [ 3, 0, 1, 1 ] );
is_deeply( \@MySTDOUT::E, [ 0, 2, 0, 0 ] );

$module->mock('epoll_wait' => sub { return(0, [
    [ fileno(*STDIN),  EPOLLIN|EPOLLHUP|EPOLLERR ],
]) });
is(loop(), 0);
is_deeply( \@MySTDIN::E,  [ 3, 0, 1, 1 ] );
is_deeply( \@MySTDOUT::E, [ 0, 2, 0, 0 ] );

Epoll_del($stdin);
Epoll_del($stdout);

#########################
### weaken references ###
#########################

$module->mock('epoll_wait' => sub { return(2, [
    [ fileno(*STDIN),  EPOLLIN  ],
    [ fileno(*STDOUT), EPOLLOUT ],
]) });

@MySTDIN::E = @MySTDOUT::E = (0, 0, 0, 0);

my $stdin1 = bless {}, 'MySTDIN';
$stdin = Epoll_add(*STDIN, $stdin1);
{
    my $stdout1 = bless {}, 'MySTDOUT';
    $stdout = Epoll_add(*STDOUT, $stdout1);
    loop();
    is_deeply( \@MySTDIN::E,  [ 1, 0, 0, 0 ] );
    is_deeply( \@MySTDOUT::E, [ 0, 1, 0, 0 ] );
}    
eval { loop(); }; die $@ if $@;
is_deeply( \@MySTDIN::E,  [ 2, 0, 0, 0 ] );
is_deeply( \@MySTDOUT::E, [ 0, 1, 0, 0 ], 'no event because of weaken ref');

lives_ok  { Epoll_del($stdin)   } q{can Epoll_del(STDIN)};
throws_ok { Epoll_del($stdout)  } qr/wrong EPOLLID/;

###########################
### weaken references 2 ###
###########################

my ($status, @events);

$stdin  = Epoll_add(*STDIN,  Suicide->new());
$stdout = Epoll_add(*STDOUT, my $manual_destroy = Suicide->new());
weaken $manual_destroy;

# check is defined/weaken CB will be returned by Epoll_wait()
$module->mock('epoll_wait' => sub { return(2, [
    [ fileno(*STDIN),  EPOLLIN|EPOLLOUT  ],
    [ fileno(*STDOUT), EPOLLIN|EPOLLOUT  ],
]) });

($status, @events) = Epoll_wait(0);
is(@events, 4);
ok(defined($_->[0]), 'CB defined')      for @events;
ok(isweak($_->[0]), 'CB weak')          for @events;
defined $_->[0] && run_callback(@$_)    for @events;
is($Suicide::Flag, 4, 'all events received');

# check for automatical Epoll_del() inside Epoll_add()
throws_ok { Epoll_add(*STDOUT, Suicide->new()) } qr/wasn't deleted/;
$manual_destroy->HUP();
lives_ok  { Epoll_add(*STDOUT, Suicide->new()) } q{can Epoll_add() for destroyed OBJECT};

# destroy self on HUP
$module->mock('epoll_wait' => sub { return(1, [
    [ fileno(*STDIN),  EPOLLHUP  ],
]) });

($status, @events) = Epoll_wait(0);
is(@events, 1);
ok(defined($_->[0]), 'CB defined')      for @events;
ok(isweak($_->[0]), 'CB weak')          for @events;
defined $_->[0] && run_callback(@$_)    for @events;
is($Suicide::Flag, 4);

# on next STDIN's event Epoll_del() must be called automatically
$module->mock('epoll_wait' => sub { return(1, [
    [ fileno(*STDIN),  EPOLLIN|EPOLLOUT  ],
]) });

($status, @events) = Epoll_wait(0);
is(@events, 0, 'no events because of automatical Epoll_del(STDIN)');

